(* Fundamental Lemma of Persistent Homology *)

Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice. 
Require Import fintype finfun bigop finset ssralg perm zmodp matrix.
Require Import persistentincidencematrices incidencematrices.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section general_case. 

Open Local Scope ring_scope.

Variable (R : ringType).

(* c is the p persistent betti number *)

Variable c : nat -> nat -> R.

(* a is beta_p^{i,j-1} - beta_p^{i,j} *)

Definition a (i j : nat) := c i j.-1 - c i j.

Definition b (i j : nat) :=
match i with
| 0 => 0%R
| S n => a n j
end.

Lemma lemma1 (k l p : nat) :
   \sum_(0<=i<k.+1) (\sum_(l.+1 <= j <p) ((a i j) - (b i j))) = \sum_(l.+1<=j<p) (a k j).
Proof.
have H : forall i, (fun _ => true) i  -> (fun i => \sum_(l.+1 <= j < p) (a i j - b i j)) i = 
         (fun i => \sum_(l.+1 <= j < p) (a i j) - \sum_(l.+1 <= j < p) (b i j)) i.
 by move => H1 /=; rewrite GRing.Theory.sumrB.
have -> : \sum_(0 <= i < k.+1) \sum_(l.+1 <= j < p) (a i j - b i j) =
          \sum_(i <- (iota 0 k.+1)) \sum_(l.+1 <= j < p) (a i j - b i j)
 by rewrite /index_iota subn0.
rewrite (@eq_bigr R 0%R _ _ _ (fun _ => true)
      (fun i => \sum_(l.+1 <= j < p) (a i j - b i j))
      (fun i => \sum_(l.+1 <= j < p) (a i j) - \sum_(l.+1 <= j < p) (b i j))
      H).
rewrite GRing.Theory.sumrB.
rewrite (@big_nat_recl _ _ _ _ (fun i => (\sum_(l.+1 <= j < p) b i j))).
rewrite /b (@big1 _ _ _ _ _ _ (fun _ : nat => 0)) // GRing.add0r.
rewrite big_nat_recr.
have -> : (\big[GRing.add_monoid R/0]_(0 <= i < k) (\sum_(l.+1 <= j < p) a i j)) = 
   \sum_(0 <= i < k) \sum_(l.+1 <= j < p) a i j by done.
have -> : (GRing.add_monoid R) (\sum_(0 <= i < k) \sum_(l.+1 <= j < p) a i j)
     (\sum_(l.+1 <= j < p) a k j) =
          (\sum_(l.+1 <= j < p) a k j) + (\sum_(0 <= i < k) \sum_(l.+1 <= j < p) a i j)
  by rewrite GRing.Theory.addrC.
rewrite -GRing.Theory.addrA.
have H1 : \sum_(0 <= i < k) \sum_(l.+1 <= j < p) a i j == \sum_(0 <= i < k) \sum_(l.+1 <= j < p) a i j 
  by done.
rewrite -GRing.subr_eq0 in H1.
by rewrite (eqP H1) GRing.addr0.
Qed.

Lemma subr_sub (m n p : R) : m - n - p = m - (n + p).
Proof.
apply/eqP.
by rewrite !GRing.subr_eq -GRing.addrA [p+n]GRing.addrC GRing.subrK.
Qed.

Lemma lemma2_aux (k l p : nat) (H : l < p) :
  \sum_(l<=j<p) ((c k j.-1) - (c k j)) = (c k l.-1) - (c k p.-1).
Proof.
rewrite GRing.Theory.sumrB big_ltn // (@big_add1 _ _ _ _ _ (fun _ => true)).
rewrite -{2 3}[l]add0n !big_addn.
have [r H1] : exists r, (p - l)%N = r.+1
  by exists (p - l).-1; rewrite -subn_gt0 in H; rewrite (ltn_predK H).
rewrite H1 big_nat_recr.
have -> :  (GRing.add_monoid R)
     (\big[GRing.add_monoid R/0]_(0 <= i < r) c k (i + l))
     (c k (r + l)) = \sum_(0<=i<r) (c k (i+l)) + (c k (r + l)) by done.
have H3 : l <= p by rewrite leq_eqVlt; apply/orP; right.
have -> : r = (p - l).-1%N.
  by rewrite H1 -pred_Sn. 
rewrite -GRing.addrA -subr_sub.
have -> : (p.-1 - l)%N = (p - l).-1.
  by rewrite -subn1 -subnDA addnC addn1 subnS.
have H4 : \sum_(0 <= i < (p - l).-1) c k (i + l) == 
          \sum_(0 <= i < (p - l).-1) c k (i + l)
  by done.
rewrite -GRing.subr_eq0 in H4; rewrite (eqP H4) GRing.sub0r.
have -> : ((p - l).-1 + l)%N = p.-1.
  by rewrite -subnS addnC addnBA -addn1 ?subnDl ?subn1 ?addn1.
by done.
Qed.

Lemma lemma2 (k l p : nat) (H : l < p) :
  \sum_(l.+1<=j<p.+1) ((c k j.-1) - (c k j)) = (c k l) - (c k p).
Proof.
have H1 : l.+1 < p.+1 by done. 
by rewrite (lemma2_aux _ H1) -!pred_Sn.
Qed.

Lemma lemma2' (k l p : nat) (H : l = p) :
  \sum_(l.+1<=j<p.+1) ((c k j.-1) - (c k j)) = (c k l) - (c k p).
Proof.
rewrite H.
move : (eq_refl (c k p)); rewrite -GRing.subr_eq0; move/eqP => ->. 
rewrite /index_iota /=.
have : p.+1 <= p.+1 by done.
by rewrite -subn_eq0; move/eqP => -> /=; rewrite big_nil. 
Qed.

Lemma lemma3 (k l p : nat) (H : l <= p) :
  \sum_(0<=i<k.+1) (\sum_(l.+1 <= j <p.+1) ((a i j) - (b i j))) =
  (c k l) - (c k p).
Proof.
move : H; rewrite leq_eqVlt; move/orP => [H|H].
  by move : H; move/eqP => H; rewrite lemma1 /a lemma2'.
by rewrite lemma1 /a lemma2. 
Qed.

End general_case. 

Section definitions.

Open Local Scope ring_scope.
Import GRing.Theory.

Variable (V:finType) (f: (seq {set (simplex V)})) (n j p:nat).
Hypothesis f_is_filtration : filtration f.
Hypothesis j_is_in_filtration :  j < size f.
Hypothesis j_leq_p_is_in_filtration : j<= p < size f.

Require Import ssrint.

Definition p_persistent_n_betti_K_j : int :=
   (p_persistent_n_betti_K_i f n j p)%:Z.

End definitions.

Section fundamental_lemma.

Open Local Scope ring_scope.
Import GRing.Theory.

Variable (V:finType) (f: (seq {set (simplex V)})) (n : nat).

Definition live_before_i_die_j (i j : nat) :=
  a (p_persistent_n_betti_K_j f n) i j.  

Definition live_before_i_1_die_j (i j : nat) :=
  b (p_persistent_n_betti_K_j f n) i j.

Definition mu (i j : nat) :=
  (live_before_i_die_j i j - live_before_i_1_die_j i j)%Z.

Lemma fundamental_lemma_persistent_homology (k l : nat) (H : l <= (size f)) : 
  \sum_(0<=j<k.+1) (\sum_(l.+1 <= p < (size f).+1) (mu j p))%Z =
  ((p_persistent_n_betti_K_j f n k l) - 
  (p_persistent_n_betti_K_j f n k (size f)))%Z.
Proof.
by rewrite (lemma3 (p_persistent_n_betti_K_j f n) _ H).
Qed.

End fundamental_lemma.








