Require Import ssreflect ssrfun ssrbool eqtype ssrnat seq choice fintype.
Require Import perm bigop finfun tuple ssralg matrix mxalgebra zmodp vector.

Require Import ssrcomplements seqmatrix cssralg.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Import GRing.Theory.

Section KerMxleft.

Local Open Scope ring_scope.

Variable (K : fieldType).

Fixpoint ker_step (m n : nat) {struct n} :=
  match n return
     'M[K]_(1+m,1+n) -> ('M[K]_(1+n) * 'M[K]_(m,n) * K * 'cV[K]_m)+
                        ('M[K]_(m, 1 + n)) with
  | S n' => fun (M: 'M[K]_(1 + _, 1 + _)) =>
     let: a := M 0 0 in
       if a == 0 then
         match ker_step _ _ (rsubmx M) with
           | inl (((P,R),a),c) =>
               (* let: P' := tperm_mx 0 1 *m block_mx (1%:M: 'M_1) 0 0 P *)
               let: P' := xcol 1 0 (block_mx (1%:M: 'M_1) 0 0 P)
               in inl (((P',row_mx (dlsubmx M) R),a),c)
           | inr R => inr (row_mx (dlsubmx M) R)
         end
       else
         let R := drsubmx M - (a^-1) *: (dlsubmx M *m ursubmx M) in
         let D := block_mx 1%:M (-(a^-1)*: (ursubmx M)) 0 1%:M in
           inl (((D,R),a),dlsubmx M)
  | _ => fun (M: 'M[K]_(1 + m,1 + 0)) =>
           let: a := M 0 0 in
             if a == 0 then
               inr (dsubmx M)
             else
               inl (((1%:M,0),a),dsubmx M)
   end.

Lemma xcol_tool m n (M: 'M[K]_(m, n)) (c1 c2: 'cV[K]_m) :
  xcol 1 0 (row_mx c1 (row_mx c2 M)) = row_mx c2 (row_mx c1 M).
Proof.
apply/matrixP => i j; rewrite !mxE; symmetry.
case: splitP => jj.
  rewrite !ord1 => hjj {jj}.
  have -> : j = 0 by apply/ord_inj.
  rewrite tpermR {j hjj}.
  case: splitP => x; first by rewrite !ord1.
  rewrite !mxE.
  case: splitP => xx.
    by rewrite !ord1 => hxx {xx}.
  by move => ->.
move => hjj.
have -> : j = lift 0 jj by apply/ord_inj.
rewrite !mxE.
case: splitP => jjj.
  rewrite !ord1 => hjjj {jjj}.
  have -> : jj = 0 by apply/ord_inj.
  have -> : lift 0 0 = 1 by move => ?; apply/ord_inj.
  rewrite tpermL.
  by case: splitP => x // _; rewrite !ord1 {x}.
move => hjjj.
have -> : jj = lift 0 jjj by apply/ord_inj.
rewrite tpermD //.
case: splitP => x; first by rewrite !ord1.
rewrite !mxE.
case: splitP => y.
  rewrite !ord1 => hx {y}.
  by have -> : x = 0 by apply/ord_inj.
move => hx.
have -> : x = lift 0 y by apply/ord_inj.
rewrite /= /bump !leq0n => /eqP.
rewrite !eqSS => /eqP hy.
by have -> : jjj = y by apply/ord_inj.
Qed.


Lemma tperm01_tool m n (M: 'M[K]_(m, n)) (c1 c2: 'cV[K]_m) :
  row_mx c1 (row_mx c2 M) *m tperm_mx 1 0 = row_mx c2 (row_mx c1 M).
Proof.
by rewrite -xcolE xcol_tool.
Qed.

Lemma ker_stepP : forall n m (M: 'M[K]_(1 + m, 1 + n)),
   match ker_step M with
       | inl (((P,R),a),c) => [/\ P \in unitmx, a != 0 &
                                  M *m P = block_mx a%:M 0 c R]
       | inr R => M = col_mx 0 R
   end.
Proof.
elim => [ | n hi] m M /=.
  case: ifP => hM00.
    rewrite -{1}[M]vsubmxK.
    by have -> : usubmx M = 0 by
      apply/colP => i; rewrite !mxE ord1 lshift0 (eqP hM00).
  rewrite hM00 unitmx1 mulmx1; split => //.
  rewrite -{1}[M]vsubmxK.
  have -> : usubmx M = (M 0 0)%:M
      by apply/rowP => i; rewrite !mxE ord1 lshift0.
  rewrite block_mxEv; f_equal.
   apply/rowP => i; rewrite !mxE !ord1.
    by case: splitP => z _ //; rewrite !ord1 !mxE.
  apply/colP => i; rewrite !mxE rshift1.
  by case: splitP => z _ //; rewrite !ord1 !mxE rshift1.
case: ifP => hM00; last first.
  rewrite hM00; split => //.
    by rewrite unitmxE det_ublock !det1 !mul1r unitr1.
  rewrite -{1}(submxK M) mulmx_block !mulmx1 !mulmx0 !addr0.
  have -> : ulsubmx M = (M 0 0)%:M
   by apply/rowP => i; rewrite !mxE !ord1 !lshift0.
  f_equal.
    rewrite mul_scalar_mx scalerA mulrN mulfV.
      by rewrite scaleN1r addrC subrr.
    by rewrite hM00.
  by rewrite addrC -scalemxAr scaleNr.
move: (hi _ (rsubmx M)).
have dl : dsubmx (lsubmx M) = dlsubmx M by apply/colP=> i; rewrite !mxE.
case: ker_step => [[[[P R] a] c] [h1 h2 h3] | R h ].
  split => //.
    by rewrite xcolE unitmx_mul unitmx_perm unitmxE det_ublock det1 mul1r
               -unitmxE h1.
  rewrite xcolE -{1}[M]hsubmxK mulmxA (@mul_row_block _ _ 1 _ 1 _) mulmx1 !mulmx0
    addr0 add0r h3 block_mxEh.
  rewrite tperm01_tool block_mxEh; f_equal.
  rewrite -{1}[lsubmx M]vsubmxK -[0: 'rV[K]_(1 + n)]row_mx0 -block_mxEh.
  have -> : usubmx (lsubmx M) = 0
    by apply/rowP => i; rewrite !mxE !ord1 !lshift0 (eqP hM00).
  by rewrite block_mxEv dl.
rewrite -{1}[M]hsubmxK h -{1}[lsubmx M]vsubmxK -block_mxEh.
have -> : usubmx (lsubmx M) = 0
  by apply/rowP => i; rewrite !mxE !ord1 !lshift0 (eqP hM00).
by rewrite block_mxEv row_mx0 dl.
Qed.

Fixpoint ker_dep (m n:nat) {struct m} :=
  match n,m return 'M[K]_(m,n) -> {p: nat & 'M[K]_(p,m)} with
      | S n', S m' => fun (M: 'M[K]_(1 + _, 1 + _)) =>
         match ker_step M with
           | inl (((P,R),a),c) =>
               let (q,Y) := ker_dep _ _ R in
                 existT (fun p => 'M[K]_(p,1+m'))
                        q (row_mx (- (a^-1) *: (Y *m c)) Y)
           | inr R => let (q,Y) := ker_dep _ _ R in
                 existT (fun p => 'M[K]_(p,1+m'))
                        (q.+1) (block_mx (1%:M: 'M_1) 0 0 Y)
         end
      | O,_ => fun _ => existT (fun p => 'M[K]_(p,m)) m (1%:M)
      | _,_ => fun _ => existT (fun p => 'M[K]_(p,0)) O 0
  end.

Definition ker (m n : nat) (M : 'M[K]_(m,n)) := projT2 (ker_dep M).


Lemma rank0M (n m : nat) (M : 'M[K]_(n,m)) :
  \rank (col_mx (0: 'rV[K]_m) M) = \rank M.
Proof.
apply (eqmx_trans (eqmx_sym (@addsmxE _ 1 n m 0 M))).
by rewrite addsmxC; apply: addsmx0.
Qed.

Lemma trmx_eq0 (n m : nat) (M : 'M[K]_(n,m)) : (M == 0) = (M^T == 0).
Proof.
apply/eqP/eqP=> [->|h]; first by rewrite trmx0.
by rewrite -[M]trmxK h trmx0.
Qed.

Lemma trmx_neq0 (n m : nat) (M : 'M[K]_(n,m)) : (M != 0) = (M^T != 0).
Proof.
apply/negP/negP=> h1 h2; apply: h1; first by rewrite trmx_eq0.
by rewrite -trmx_eq0.
Qed.

Lemma rankaMc : forall (n m:nat) (M: 'M[K]_(n,m)) (a:K) (c:'cV[K]_n),
 a != 0 -> \rank (block_mx (a%:M) 0 c M) = (1 + \rank M)%N.
Proof.
move => n m M a c ha.
have: ((block_mx a%:M 0 c M)^T :=: (col_mx a%:M c)^T + (col_mx 0 M)^T)%MS.
  rewrite tr_block_mx !tr_col_mx /block_mx.
  apply eqmx_sym.
  by apply addsmxE.
have hC: (col_mx a%:M c) != 0.
  apply/negP; move/eqP => h.
  move/negP : ha; apply.
  move/colP : h => h.
  move: (h 0); rewrite !mxE.
  case: splitP => i hi //.
  by rewrite [i]ord1 !mxE /= mulr1n => ->.
rewrite -mxrank_tr => ->.
rewrite mxrank_disjoint_sum.
  rewrite rank_rV mxrank_tr rank0M.
  by rewrite -trmx_neq0 hC.
rewrite !tr_col_mx !tr_scalar_mx trmx0.
apply/eqP.
apply/rowV0P => v.
rewrite sub_capmx; case/andP.
case/sub_rVP => b hb.
case/submxP => B hB.
have : v 0 0 = 0 /\ v 0 0 = b * a.
  split.
    rewrite hB !mxE big1 // => i _.
    rewrite !mxE.
    case: splitP => j //.
    by rewrite !mxE /= mulr0.
  rewrite hb !mxE.
  case: splitP => j //.
  by rewrite [j]ord1 !mxE /= mulr1n.
rewrite {3}hb; case => ->.
rewrite -[0](@mul0r _ a). move/(mulIf ha) <-.
by rewrite scale0r.
Qed.

Lemma rankaMl : forall (n m:nat) (M: 'M[K]_(n,m)) (a:K) (l:'rV[K]_m),
 a != 0 -> \rank (block_mx (a%:M) l 0 M) = (1 + \rank M)%N.
Proof.
move => n m M a l ha.
by rewrite -mxrank_tr tr_block_mx trmx0 tr_scalar_mx rankaMc // mxrank_tr.
Qed.


Lemma ker_dim : forall m n (M: 'M[K]_(m,n)),
  projT1 (ker_dep M) = (m - \rank M)%nat.
Proof.
elim => [ | m hi] [ | n] //=.
rewrite [m.+1]/(1 + m)%N [n.+1]/(1 + n)%N => M.
move: (ker_stepP M).
case: ker_step => [[[[P R] a] c] [h1 h2 h3]| R]. 
  move : (hi _ R).
  case: ker_dep => [p Y] /= ->.
  have -> : \rank M = (1 + \rank R)%N.
    have hP : row_free P by rewrite row_free_unit.
    rewrite -(mxrankMfree M hP) h3 rankaMc //.
  by rewrite [(1 + _)%N]addnC -addn1 subnDr.
move => hM.
move: (hi _ R).
case: ker_dep => [p Y] /= ->.
have -> : \rank M = \rank R  by rewrite hM rank0M.
by rewrite subSn // rank_leq_row.
Qed.

Definition kernel m n (M: 'M[K]_(m,n)) : 'M[K]_(m - \rank M,m) :=
    castmx (ker_dim M,erefl _) (ker M).

Lemma tool m n (M: 'M[K]_(m,n)) (P: 'M[K]_n) :
  P \in unitmx -> M *m P = 0 -> M = 0.
Proof.
by move => hP h; rewrite -[M]mulmx1 -(mulmxV hP) mulmxA h mul0mx.
Qed.

Lemma ker0MS : forall m n (M: 'M[K]_(m,n)), ker M *m M = 0.
               (* let (p,R) := ker M in R *m M = 0. *)
Proof.
rewrite /ker.
elim => [ | m hi] [ | n] M /=; try (by rewrite mul1mx ?flatmx0 ?thinmx0).
  by rewrite mul0mx.
move: (ker_stepP M).
case: ker_step => [[[[P R] a] c] [h1 h2 h3]| R -> {M}].
  case: ker_dep (hi _ R) => [p Y] hY.
  apply/(tool h1).
  rewrite -mulmxA h3 (@mul_row_block _ _ 1 _ 1) mulmx0 add0r hY mul_mx_scalar.
  by rewrite scalerA mulrN (mulfV h2) scaleN1r addrC subrr row_mx0.
case: ker_dep (hi _ R) => [p Y] hY.
rewrite block_mxEh (@mul_row_col _ _ 1) mulmx0 add0r (@mul_col_mx _ 1).
by rewrite mul0mx hY col_mx0.
Qed.


Lemma castmx_mul__ (R : ringType)
  (m n p p': nat) (ep : p = p')
  (M : 'M[R]_(m, n)) (N : 'M[R]_(n, p)) :
  castmx (erefl _, ep) (M *m N) = M *m castmx (erefl _, ep) N.
Proof. by  case: p' / ep. Qed.

Lemma ker_row_free_ : forall m n (M: 'M[K]_(m,n)),
  let (p,Y) := ker_dep M in row_free Y.
Proof.
elim => [| m hi][|n]//=.
  move => _; apply/row_freeP.
  exists 1%:M; by rewrite mul1mx.
rewrite [m.+1]/(1 + m)%N [n.+1]/(1 + n)%N => M.
move: (ker_stepP M).
case: ker_step => [[[[P R] a] c] [h1 h2 h3]| R ].
  move: (hi _ R).
  case: ker_dep => p Y /row_freeP [B hB].
  apply/row_freeP.
  exists (col_mx (0: 'rV_p) B).
  by rewrite (@mul_row_col _ p 1) mulmx0 add0r hB.
move => hM.
move: (hi _ R).
case: ker_dep => p Y /row_freeP [B hB].
apply/row_freeP.
rewrite [p.+1]/(1 + p)%N.
exists (block_mx (1%:M: 'M_1) 0 0 B).
rewrite (@mulmx_block _ 1 _ 1 _ 1).
by rewrite !mul1mx !mul0mx !addr0 !mulmx0 !add0r hB -scalar_mx_block.
Qed.

Lemma ker_row_free : forall m n (M: 'M[K]_(m,n)), row_free (ker M).
Proof.
rewrite /ker => m n M.
move: (ker_row_free_ M).
by case: ker_dep => [p Y] /=.
Qed.


Lemma castmx_mul (R : ringType)
  (m m' n p p': nat) (em : m = m') (ep : p = p')
  (M : 'M[R]_(m, n)) (N : 'M[R]_(n, p)) :
  castmx (em, ep) (M *m N) = castmx (em, erefl _) M *m castmx (erefl _, ep) N.
Proof. by case: m' / em; case: p' / ep. Qed.

Lemma kernel_row_free : forall m n (M: 'M[K]_(m,n)), row_free (kernel M).
Proof.
rewrite /kernel => m n M.
case/row_freeP : (ker_row_free M) => B hB.
apply/row_freeP.
exists (castmx (erefl _,ker_dim M) B).
rewrite -castmx_mul hB.
by apply/matrixP => i j; rewrite !mxE castmxE !mxE /=.
Qed.

Lemma ker_base : forall m n p (M: 'M[K]_(m,n)) (X: 'M[K]_(p,m)),
  (* X *m M = 0 -> let (q,Z) := ker M in exists Y, X = Y *m Z. *)
  X *m M = 0 -> exists Y, X = Y *m ker M.
Proof.
rewrite /ker.
elim => [ | m hi] [ | n] p //=;
  try (by move => M X _; exists X; rewrite mulmx1).
  by move => M X _; exists X; rewrite thinmx0 mulmx0.
rewrite [m.+1]/(1 + m)%N [n.+1]/(1 + n)%N => M X hX.
move: (ker_stepP M).
case: ker_step => [[[[P R] a] c] [h1 h2 h3] | R].
  have : X *m (M *m P) = 0 by rewrite mulmxA hX mul0mx.
  rewrite h3 -{1}[X]hsubmxK (@mul_row_block _ _ 1 _ 1) mulmx0 add0r.
  rewrite -{1}[0]row_mx0 => /eq_row_mx [hl hr].
  case: ker_dep (hi _ _ R (rsubmx X) hr) => q Z /= [Y hY].
  exists Y.
  rewrite (@mul_mx_row _ _ _ 1) -hY -{1}[X]hsubmxK scaleNr mulmxN scalemxAr.
  rewrite mulmxA -hY -scalemxAr.
  move/eqP: hl; rewrite addrC addr_eq0 => /eqP ->.
  by rewrite scalerN opprK mul_mx_scalar scalerA mulVf // scale1r.
move => hM; move : hX; rewrite hM {M hM}.
rewrite -{1}[X]hsubmxK (@mul_row_col _ p 1) mulmx0 add0r => hR.
case: ker_dep (hi _ _ R (rsubmx X) hR) => q Y [Z hZ].
exists (row_mx (lsubmx X) Z).
by rewrite (@mul_row_block _  _ 1 _ 1) !mulmx0 addr0 add0r -hZ mulmx1 hsubmxK.
Qed.

Lemma sub_kerP m n p (M : 'M[K]_(m,n)) (X : 'M[K]_(p,m)) :
  reflect (X *m M = 0) (X <= ker M)%MS.
Proof.
apply: (iffP submxP) => [[Y ->]|h0]; first by rewrite -mulmxA ker0MS mulmx0.
by apply/ker_base.
Qed.

End KerMxleft.

Section ker_seqmx.

Variable (K : fieldType) (CK : cunitRingType K).

Fixpoint ker_seqmx_step (m n : nat) (M : seqmatrix CK)
  : (seqmatrix CK * CK * seqmatrix CK) + (seqmatrix CK) :=
  match n with
  | S n' =>
      let a := M 0 0 in
      if a == zero CK then
         match ker_seqmx_step m n' (rsubseqmx 1 M) with
           | inl ((R,a),c) => inl ((row_seqmx (dlsubseqmx 1 1 M) R,a),c)
           | inr R => inr (row_seqmx (dlsubseqmx 1 1 M) R)
         end
      else
         let R := subseqmx (drsubseqmx 1 1 M) (scaleseqmx (cinv a)
                       (mulseqmx 1 n (dlsubseqmx 1 1 M) (ursubseqmx 1 1 M))) in
           inl ((R,a),dlsubseqmx 1 1 M)
  | _ =>
      let a := M 0 0 in
      if a == zero CK then inr (dsubseqmx 1 M) else inl ((seqmx0 CK m n,a),dsubseqmx 1 M)
  end.

Lemma ker_seqmx_stepE : forall n m (M : 'M[K]_(m.+1,n.+1)),
  match ker_step M, ker_seqmx_step m n (seqmx_of_mx CK M) with
  | inl (((_,M),a),c), inl ((M',a'),c') => [/\ seqmx_of_mx _ M = M', trans a = a' & seqmx_of_mx _ c = c']
  | inr R, inr R' => seqmx_of_mx _ R = R'
  | _,_ => false
  end.
Proof.
elim=> [m M /=|n ih m].
  case: ifP=> h.
    have -> : (seqmx_of_mx CK M) 0 0 == zero CK
      by rewrite (@seqmxE _ _ _ _ _ ord0 ord0) (eqP h) zeroE.
    by rewrite dsubseqmxE.
   have -> : ((seqmx_of_mx CK M) 0 0 == zero CK) = false
     by rewrite (@seqmxE _ _ _ _ _ ord0 ord0) trans_eq0 h.
   by rewrite seqmx0E (@seqmxE _ _ _ _ _ ord0 ord0) dsubseqmxE.
rewrite [m.+1]/(1 + m)%N [n.+2]/(1 + (1 + n))%N => M /=.
move: (ih _ (rsubmx M)).
rewrite {ih}.
case: ifP => h /=.
  case: ker_step=> [[[[_ N a c]]]|R]; rewrite rsubseqmxE.
    have -> : (seqmx_of_mx CK M) 0 0 == zero CK
      by rewrite (@seqmxE _ _ _ _ _ ord0 ord0) (eqP h) zeroE.
    case: ker_seqmx_step=> [[[N' a' c']]|] // [<- -> ->].
    by rewrite dlsubseqmxE row_seqmxE.
  have -> : (seqmx_of_mx CK M) 0 0 == zero CK
    by rewrite (@seqmxE _ _ _ _ _ ord0 ord0) (eqP h) zeroE.
  case: ker_seqmx_step=> [[[N' a' c']]|R'] // <-.
  by rewrite dlsubseqmxE row_seqmxE.
have -> : ((seqmx_of_mx CK M) 0 0 == zero CK) = false
  by rewrite (@seqmxE _ _ _ _ _ ord0 ord0) trans_eq0 h.
move=> ih.
rewrite (@seqmxE _ _ _ _ _ ord0 ord0) dlsubseqmxE.
by rewrite ursubseqmxE mulseqmxE drsubseqmxE -cinvE scaleseqmxE subseqmxE.
Qed.

Fixpoint ker_seqmx (m n:nat) (M : seqmatrix CK) {struct m} : seqmatrix CK :=
  match n,m with
    | S n', S m' =>
         match ker_seqmx_step m' n' M with
           | inl ((R,a),c) =>
               let Y := ker_seqmx m' n' R in
               row_seqmx (scaleseqmx (opp (cinv a))
                         (* A trick to avoid the problem with zero size in mulseqmxE... *)
                         (if m' == 0 then seqmx0 _ (size Y) 1 else mulseqmx m' 1 Y c)) Y
           | inr R => let Y := ker_seqmx m' n R in
                      block_seqmx (seqmx1 _ 1) (seqmx0 _ 1 m') (seqmx0 _ (size Y) 1) Y
         end
      | O,_ => seqmx1 _ m
      | _,_ => seqmx0 _ O O
  end.


Lemma ker_seqmxE : forall m n (M : 'M[K]_(m,n)),
  seqmx_of_mx _ (ker M) = ker_seqmx m n (seqmx_of_mx _ M).
Proof.
rewrite /ker; elim => [m M | m ih n M].
  rewrite thinmx0 flatmx0.
  by case: m M=> [|m] M; rewrite seqmx0E.
case: n ih M => [|n ih M]; first by rewrite /ker /= seqmx1E.
move: M (@ker_seqmx_stepE n m M).
rewrite [m.+1]/(1 + m)%N [n.+1]/(1 + n)%N=> M hM /=.
move: hM.
case: ker_step=> [[[[_ N a c]]]|R].
  case: ker_seqmx_step=> [[[N' a' c']]|R'] // [<- <- <-].
  rewrite -ih.
  case: ker_dep=> /= p Y.
  have -> : (if m == 0 then seqmx0 _ (size (seqmx_of_mx _ Y)) 1
                       else mulseqmx m 1 (seqmx_of_mx CK Y) (seqmx_of_mx CK c))
            = seqmx_of_mx CK (Y *m c).
   move=> R {ih M N N' c' a'}.
   case: m Y c => [|m] Y c; last by rewrite mulseqmxE.
   by rewrite !thinmx0 mul0mx !seqmx0E /seqmx0 /const_seqmx size_nseq.
  by rewrite -cinvE -oppE scaleseqmxE row_seqmxE.
case: ker_seqmx_step=> [|R'] // <- /=.
rewrite -ih.
case: ker_dep=> p Y /=.
have h0 : forall n, [:: nseq n (zero CK)] = (seqmx0 _ 1 n) by [].
by rewrite !h0 -!seqmx0E -seqmx1E size_seqmx block_seqmxE.
Qed.

End ker_seqmx.






