Require Import ssreflect ssrbool ssrfun eqtype ssrnat seq div choice fintype.
Require Import finfun bigop finset ssralg perm zmodp matrix fingroup mxalgebra.
Require Import path vector seqmatrix boolF2 incidencematrices persistenthomology.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section filtration.

(* V is the vertex set *)
Variable V : finType.

(* Filtration definifition *)
Definition filtration (f : seq {set simplex V}) :=
    (forall x, x \in f -> simplicial_complex x) /\
    (forall i j, i <= j -> j < size f -> (nth set0 f i) \subset (nth set0 f j)).

End filtration.

Section PersistenceFiltration.

Section inclusion_matrix.

Open Local Scope ring_scope.

Variable V : finType.
Variables m n : nat.

(* Top and Left are two enumerated finite sets of simplices *)
Variables Left Top : seq (simplex V).

Definition inclusionMatrix :=
  \matrix_(i < m, j < n)
    if (nth set0 Left i == nth set0 Top j) then 1 else 0:'F_2.

Definition projectionMatrix :=
  \matrix_(i < n, j < m)
    if (nth set0 Top i == nth set0 Left j) then 1 else 0:'F_2.

End inclusion_matrix.



Section nth_inclusion_mx.

Open Local Scope ring_scope.
Import GRing.Theory.

Variable (V:finType) (f: (seq {set (simplex V)})) (n j p:nat).
Hypothesis f_is_filtration : filtration f.
Hypothesis j_is_in_filtration :  j < size f.
Hypothesis j_leq_p_is_in_filtration : j <= p < size f.

Definition n_simplices_k k := n_simplices (nth set0 f k) n.

Lemma n_k_simplices_subset j0 l0 : j0 < size f -> l0 < size f -> j0 <= l0 -> n_simplices_k j0 \subset n_simplices_k l0.
Proof.
move => H1 H2 H3.
apply/subsetP => x.
rewrite !inE; move/andP => [H4 H5].
apply/andP; split; last by done.
move : f_is_filtration. rewrite /filtration; move => [H6 H7].
by move : (H7 j0 l0 H3 H2); move/subsetP => H8; apply : (H8 x H4).
Qed.

Definition inclusion_mx :=
  inclusionMatrix (Vector.dim (incidencematrices.V1 (nth set0 f j) n.+1))
                  (Vector.dim (incidencematrices.V2 (nth set0 f p) n))
                  (enum (n_simplices_k j))
                  (enum (n_simplices_k p)).

Definition projection_mx :=
  projectionMatrix (Vector.dim (incidencematrices.V1 (nth set0 f j) n.+1))
                  (Vector.dim (incidencematrices.V2 (nth set0 f p) n))
                  (enum (n_simplices_k j))
                  (enum (n_simplices_k p)).

Lemma exist_1_inclusion_mx j0 : exists l, inclusion_mx j0 l = 1%R.
Proof.
rewrite /inclusion_mx /inclusionMatrix /=.
have H : (index (nth set0 (enum (n_simplices_k j)) j0)
            (enum (n_simplices_k p))) <
               Vector.dim (incidencematrices.V2 (nth set0 f p) n); last first.
  exists (Ordinal H).
  rewrite mxE.
  have -> : nth set0 (enum (n_simplices_k j)) j0 == nth set0 (enum (n_simplices_k p)) (Ordinal H); last by done.
  rewrite nth_index; first by done.
  have H1 :  nth set0 (enum (n_simplices_k j)) j0 \in (enum (n_simplices_k j)).
    rewrite mem_nth //; move : (ltn_ord j0); rewrite /Vector.dim /V1 /= -cardE.
    move => H1; rewrite -[X in _ < X]mul1n; move : H1.
    by rewrite /n_simplices_k /n_simplices /n_1_simplices /=.
  move : f_is_filtration. rewrite /filtration; move => [H2 H3].
  move : j_leq_p_is_in_filtration; move/andP => [H4a H4b].
  move : (n_k_simplices_subset j_is_in_filtration H4b H4a).
  move/subsetP => H5.
  rewrite mem_enum.
  rewrite mem_enum in H1.
  by move : (H5 _ H1).
rewrite /Vector.dim /V2 /=.
have -> : (1 * #|n_simplices (nth set0 f p) n|)%N = size  (enum (n_simplices_k p))
  by rewrite cardE /n_simplices_k mul1n.
rewrite index_mem.
have H1 : nth set0 (enum (n_simplices_k j)) j0 \in (enum (n_simplices_k j)).
    rewrite mem_nth //; move : (ltn_ord j0); rewrite /Vector.dim /V1 /= -cardE.
    move => H1; rewrite -[X in _ < X]mul1n; move : H1.
    by rewrite /n_simplices_k /n_simplices /n_1_simplices /=.
  move : f_is_filtration. rewrite /filtration; move => [H2 H3].
  move : j_leq_p_is_in_filtration; move/andP => [H4a H4b].
  move : (n_k_simplices_subset j_is_in_filtration H4b H4a).
  move/subsetP => H5.
  rewrite mem_enum.
  rewrite mem_enum in H1.
  by move : (H5 _ H1).
Qed.

Lemma inclusion_uniq_element j0 l1 l2:
  inclusion_mx j0 l1 = 1 -> inclusion_mx j0 l2 = 1 -> l1 = l2.
Proof.
rewrite !mxE.
case:ifP; last by done.
case:ifP; last by done.
move => H1.
rewrite (eqP H1) nth_uniq ?enum_uniq //.
  by move/eqP => H2 _ _; apply:val_inj.
  move : (ltn_ord l2). rewrite /Vector.dim /V1 /=.
  rewrite -cardE /n_simplices_k /n_simplices /n_1_simplices.
  by move => H2; rewrite -[X in _ < X]mul1n.
  move : (ltn_ord l1). rewrite /Vector.dim /V1 /=.
  rewrite -cardE /n_simplices_k /n_simplices /n_1_simplices.
  by move => H2; rewrite -[X in _ < X]mul1n.
Qed.


Lemma inclusion_uniq_element_2 j1 j2 l :
  inclusion_mx j1 l = 1 -> inclusion_mx j2 l = 1 -> j1 = j2.
Proof.
rewrite !mxE.
case:ifP; last by done.
case:ifP; last by done.
move => H1.
rewrite -(eqP H1) nth_uniq ?enum_uniq //.
  by move/eqP => H2 _ _; apply:val_inj.
  move : (ltn_ord j1). rewrite /Vector.dim /V1 /=.
  rewrite -cardE /n_simplices_k /n_simplices /n_1_simplices.
  by move => H2; rewrite -[X in _ < X]mul1n.
  move : (ltn_ord j2). rewrite /Vector.dim /V1 /=.
  rewrite -cardE /n_simplices_k /n_simplices /n_1_simplices.
  by move => H2; rewrite -[X in _ < X]mul1n.
Qed.


Lemma inclusion_projection_transpose j0 l :
   inclusion_mx j0 l = projection_mx l j0.
Proof.
rewrite !mxE.
by case:ifP; rewrite eq_sym => ->.
Qed.


Lemma inclusion_projection : inclusion_mx *m projection_mx = 1%:M.
Proof.
apply/matrixP => j0 l. rewrite mxE.
rewrite /scalar_mx mxE /=.
case H : (j0==l).
rewrite (eqP H).
move : (exist_1_inclusion_mx j0) => [x Hx].
rewrite (bigD1 x) //.
  move : (eqP H) => <-; rewrite -inclusion_projection_transpose Hx /= mul1r.
  have -> : \sum_(i0 < Vector.dim (incidencematrices.V2 (nth set0 f p) n) | i0 != x)
   inclusion_mx j0 i0 * projection_mx i0 j0 = 0; last by rewrite GRing.addr0.
  rewrite big1 //.
  move => i0 Hxi0.
  have [H1|H1] : inclusion_mx j0 i0 = 0 \/ inclusion_mx j0 i0 = 1.
    rewrite !mxE; case:ifP; first by right. by left.
    by rewrite H1 GRing.mul0r.
  move : (inclusion_uniq_element Hx H1) Hxi0 => ->.
  by move/eqP.
rewrite big1 //.
move => i0 _.
rewrite -inclusion_projection_transpose.
have [H1|H1] : inclusion_mx j0 i0 = 0 \/ inclusion_mx j0 i0 = 1.
  rewrite !mxE; case:ifP; first by right. by left.
  by rewrite H1 GRing.mul0r.
have [H2|H2] : inclusion_mx l i0 = 0 \/ inclusion_mx l i0 = 1.
  rewrite !mxE; case:ifP; first by right. by left.
  by rewrite H2 GRing.mulr0.
move : (inclusion_uniq_element_2 H1 H2) H => ->.
by rewrite eqxx.
Qed.


Lemma injective_LinearApp_inclusion_mx : injective (Vector.Hom (inclusion_mx)).
Proof.
move => x y.
rewrite /fun_of_lfun unlock => /= H1.
move : (Vector.InternalTheory.r2v_inj H1).
move => H.
have H2 : Vector.InternalTheory.v2r x *m inclusion_mx *m projection_mx = Vector.InternalTheory.v2r y *m inclusion_mx *m projection_mx.
  rewrite -!mulmxA !inclusion_projection !mulmx1.
  move : H.
  rewrite /mulmxr_head //= => H.
  have : Vector.InternalTheory.v2r x *m inclusion_mx *m projection_mx =
      Vector.InternalTheory.v2r y *m inclusion_mx *m projection_mx.
    by rewrite H.
  by rewrite -!mulmxA !inclusion_projection !mulmx1.
rewrite -!mulmxA !inclusion_projection !mulmx1 in H2.
by exact : Vector.InternalTheory.v2r_inj.
Qed.

End nth_inclusion_mx.

Section persistent_incidence_mx.

Open Local Scope ring_scope.

Variable (V:finType) (f: (seq {set (simplex V)})) (n j p:nat).
Hypothesis f_is_filtration : filtration f.
Hypothesis j_is_in_filtration :  j < size f.
Hypothesis j_leq_p_is_in_filtration : j <= p < size f.

Definition p_persistent_n_homology_K_j :=
 PHomology (Vector.Hom (incidence_mx_n (nth set0 f j) n)^T)
        (Vector.Hom (incidence_mx_n (nth set0 f p) n.+1)^T)
        (Vector.Hom (inclusion_mx f n j p)).

Definition p_persistent_n_betti_K_j :=
 PBetti (Vector.Hom (incidence_mx_n (nth set0 f j) n)^T)
         (Vector.Hom (incidence_mx_n (nth set0 f p) n.+1)^T)
         (Vector.Hom (inclusion_mx f n j p)).

Definition p_persistent_n_betti_K_i :=
 PBetti_rank (incidence_mx_n (nth set0 f j) n)^T
             (incidence_mx_n (nth set0 f p) n.+1)^T
             (inclusion_mx f n j p).


Lemma p_persistent_n_betti_K_iE :
  p_persistent_n_betti_K_i =
  PBetti (Vector.Hom (incidence_mx_n (nth set0 f j) n)^T)
         (Vector.Hom (incidence_mx_n (nth set0 f p) n.+1)^T)
         (Vector.Hom (inclusion_mx f n j p)).
Proof.
rewrite /p_persistent_n_betti_K_i PBetti_rankE //.
by apply : injective_LinearApp_inclusion_mx.
Qed.

End persistent_incidence_mx.



Section executable_persistent_inclusion_mx.

Local Open Scope ring_scope.

Variable V : finType.
Variable leT : rel V.

Hypothesis tr_leT : transitive leT.
Hypothesis irr_leT : irreflexive leT.
Hypothesis t_leT : total leT.

Variable filtration_faces : seq (seq (seq V)).
Variable n j p: nat.

Hypothesis Hjsize : j < size filtration_faces.
Hypothesis Hjpsize : j<= p < size filtration_faces.

Hypothesis Hfaces : all (fun s => all (sorted leT) s) filtration_faces.
Hypothesis Hfaces_uniq : all (fun s => all uniq s) filtration_faces.


Definition ex_inclusion_mx :=
  let ex_boundaries := (ex_n_simplices (nth nil filtration_faces p) n) in
  (map (fun x => map (fun y => if x == y then true else false) ex_boundaries)
    (ex_n_simplices (nth nil filtration_faces j) n)).


Definition InM := inclusionMatrix
                     (Vector.dim (incidencematrices.V2 (sc_set (nth nil filtration_faces j)) n))
                     (Vector.dim (incidencematrices.V2 (sc_set (nth nil filtration_faces p)) n))
                     (n_simplices_seq (nth nil filtration_faces j) n)
                     (n_simplices_seq (nth nil filtration_faces p) n).


Lemma all_sorted_i : all (sorted leT) (nth [::] filtration_faces j).
Proof.
apply/allP => x H.
move:Hfaces; move/allP => H1.
have H2 : (nth nil filtration_faces j) \in filtration_faces by rewrite mem_nth //.
move : (H1 (nth nil filtration_faces j) H2).
move/allP => H3.
by move : (H3 x H).
Qed.



Lemma all_sorted_ip : all (sorted leT) (nth [::] filtration_faces p).
Proof.
apply/allP => x H.
move:Hfaces; move/allP => H1.
have H2 : (nth nil filtration_faces p) \in filtration_faces
  by move : Hjpsize; move/andP => [H2a H2b]; rewrite mem_nth //.
move : (H1 (nth nil filtration_faces p) H2).
move/allP => H3.
by move : (H3 x H).
Qed.


Lemma uniq_i j0 : j0 < size (ex_n_simplices (nth [::] filtration_faces j) n) ->  uniq (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0).
Proof.
move => H.
have H1 : (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) \in (ex_n_simplices (nth [::] filtration_faces j) n) by rewrite mem_nth.
have H2 : forall x, x \in (ex_n_simplices (nth [::] filtration_faces j) n) -> x\in  (sc (nth nil filtration_faces j)).
  move => x. rewrite /ex_n_simplices mem_filter.
  by move/andP => [_ ->].
have H3 : (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) \in sc (nth nil filtration_faces j).
  by move : (H2 (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) H1).
move : (sc_uniq tr_leT irr_leT all_sorted_i).
move/allP => H4.
by move : (H4 (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) H3).
Qed.

Lemma size_i j0 :
  j0 < size (ex_n_simplices (nth [::] filtration_faces j) n) ->
 size (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) = n.+1.
Proof.
move => H.
have : (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) \in (ex_n_simplices (nth [::] filtration_faces j) n) by rewrite mem_nth.
rewrite {2}/ex_n_simplices mem_filter; move/andP => [H1 _].
by rewrite (eqP H1).
Qed.

Lemma sorted_i j0 :
  j0 < size (ex_n_simplices (nth [::] filtration_faces j) n) ->
  sorted leT (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0).
Proof.
move => H.
have H1 : (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) \in (ex_n_simplices (nth [::] filtration_faces j) n) by rewrite mem_nth.
move : all_sorted_i; move/allP => H2.
have H3 : forall x, x \in ex_n_simplices (nth [::] filtration_faces j) n ->
 x \in  sc (nth [::] filtration_faces j)
 by move => x; rewrite /ex_n_simplices mem_filter; move/andP => [_ Hk].
move : all_sorted_i => Hi.
move : (sc_sorted tr_leT Hi).
move/allP => H4.
by move : (H4 _ (H3 _ H1)).
Qed.

Lemma sorted_ip j0 :
  j0 < size (ex_n_simplices (nth [::] filtration_faces p) n) ->
  sorted leT (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0).
Proof.
move => H.
have H1 : (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) \in (ex_n_simplices (nth [::] filtration_faces p) n) by rewrite mem_nth.
move : all_sorted_i; move/allP => H2.
have H3 : forall x, x \in ex_n_simplices (nth [::] filtration_faces p) n ->
 x \in  sc (nth [::] filtration_faces p)
 by move => x; rewrite /ex_n_simplices mem_filter; move/andP => [_ Hk].
move : all_sorted_ip => Hi.
move : (sc_sorted tr_leT Hi).
move/allP => H4.
by move : (H4 _ (H3 _ H1)).
Qed.

Lemma size_ip j0 :
  j0 < size (ex_n_simplices (nth [::] filtration_faces p) n) ->
 size (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) = n.+1.
Proof.
move => H.
have : (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) \in (ex_n_simplices (nth [::] filtration_faces p) n) by rewrite mem_nth.
rewrite {2}/ex_n_simplices mem_filter; move/andP => [H1 _].
by rewrite (eqP H1).
Qed.


Lemma uniq_ip j0 :
j0 < size (ex_n_simplices (nth [::] filtration_faces p) n) ->
uniq (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0).
Proof.
move => H.
have H1 : (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) \in (ex_n_simplices (nth [::] filtration_faces p) n) by rewrite mem_nth.
have H2 : forall x, x \in (ex_n_simplices (nth [::] filtration_faces p) n) -> x\in  (sc (nth nil filtration_faces p)).
  move => x. rewrite /ex_n_simplices mem_filter.
  by move/andP => [_ ->].
have H3 : (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) \in sc (nth nil filtration_faces p).
  by move : (H2 (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) H1).
move : (sc_uniq tr_leT irr_leT all_sorted_ip).
move/allP => H4.
by move : (H4 (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) j0) H3).
Qed.



Lemma sorted_leq x1 x2 (s2:seq V): leT x1 x2 -> sorted leT (x2::s2) ->
  x1 \notin s2.
Proof.
elim : s2 => [|hs2 ts2 IH] // lx1x2 Hs.
have H : sorted leT (x2::ts2).
  move : Hs; rewrite {1}/sorted /=; move/andP => [H1]; move/pathP => H2.
  apply/(pathP x2) => i0 H0 /=.
  have Hj : forall j, 0 < j ->
    (nth x2 (x2 :: ts2) j) = (nth x2 (hs2 :: ts2) j).
    move => j1 j0.
    by have [k -> //] : exists k, j1 = k.+1
      by exists j1.-1; rewrite (prednK j0).
  elim : i0 H0 => //; last first.
    move => n0 IH0.
    rewrite (Hj n0.+1) // => H.
    by exact : (H2 x2 n0.+1 H).
  have H : leT (nth x2 (x2::ts2) 0) (nth x2 (hs2::ts2) 0) by done.
  move => H0.
  move : (H2 x2 0%N H0) => H0'.
  by exact : (tr_leT H H0').
have H1 :  leT x2 hs2.
  by move : Hs; rewrite {1}/sorted /=; move/andP => [H1].
rewrite inE negb_or (IH lx1x2 H) Bool.andb_true_r.
case H2 : (x1 == hs2); last by apply/eqP.
by move : (tr_leT lx1x2 H1); rewrite (eqP H2); move : (irr_leT hs2) => ->.
Qed.


Lemma sorted_in x1 x2 (s2:seq V): x2 \in s2 -> sorted leT (x1::s2) ->
  leT x1 x2.
Proof.
elim : s2 => [|hs2 ts2 IH] // lx1x2 Hs.
have H : sorted leT (x1::ts2).
  move : Hs; rewrite {1}/sorted /=; move/andP => [H1]; move/pathP => H2.
  apply/(pathP x2) => i0 H0 /=.
  have Hj : forall j, 0 < j ->
    (nth x2 (x1 :: ts2) j) = (nth x2 (hs2 :: ts2) j).
    move => j1 j0.
    by have [k -> //] : exists k, j1 = k.+1
      by exists j1.-1; rewrite (prednK j0).
  elim : i0 H0 => //; last first.
    move => n0 IH0.
    rewrite (Hj n0.+1) // => H.
    by exact : (H2 x2 n0.+1 H).
  have H : leT (nth x2 (x1::ts2) 0) (nth x2 (hs2::ts2) 0) by done.
  move => H0.
  move : (H2 x2 0%N H0) => H0'.
  by exact : (tr_leT H H0').
move : lx1x2; rewrite inE; move/orP => [H1|H1].
  have H2 :  leT x1 hs2
    by move : Hs; rewrite {1}/sorted /=; move/andP => [Hk].
  by rewrite (eqP H1).
by exact : (IH H1 H).
Qed.

Lemma sorted_exists : forall (s1 s2 : seq V), size s1 = size s2 ->
  (sorted leT s1) -> (sorted leT s2) -> s1 <> s2 ->
  exists x, x \in s1 /\ x \notin s2.
Proof.
elim=> [|x1 s1 IHs1] [|x2 s2] // H1 H2 H3 H4.
have H1' : sorted leT s1
  by move : H2; rewrite {1}/sorted => H; apply : (path_sorted H).
have H2' : sorted leT s2
  by move : H3; rewrite {1}/sorted => H; apply : (path_sorted H).
 have H3' : size s1 = size s2.
  exact : (eq_add_S _ _ H1).
case H : (s1 == s2).
  case H0 : (x1 == x2).
   by  move : H4; rewrite (eqP H) (eqP H0).
  exists x1; split; rewrite inE.
    by apply/orP; left.
  rewrite negb_or; apply/andP; split.
    by apply/eqP; move : H0; move/eqP.
  rewrite -(eqP H).
  move : (sorted_uniq tr_leT irr_leT H2).
  by rewrite cons_uniq; move/andP => [-> _].
move : H; move/eqP => H.
move : (IHs1 _ H3' H1' H2' H) => [x [Hx1 Hx2]].
move : (t_leT x1 x2); move/orP => [Ht1|Ht1].
  exists x1; split; rewrite inE.
    by apply/orP; left.
  rewrite negb_or; apply/andP; split.
    case H0 : (x1 == x2); last by move : H0;  move/eqP.
    move : Ht1; rewrite (eqP H0).
    by move : (irr_leT x2) => ->.
 by exact : (sorted_leq Ht1 H3).
exists x; split; rewrite inE.
  by apply/orP; right.
rewrite negb_or; rewrite Hx2 Bool.andb_true_r.
move : (sorted_in Hx1 H2) => Hs.
move : (tr_leT Ht1 Hs) => Ht.
case Hf : (x == x2); last by apply/eqP.
by move : Ht; rewrite (eqP Hf); move : (irr_leT x2) => ->.
Qed.


Lemma set0_setin (s : seq V) : uniq s -> [set t in s] = set0 -> s = nil.
Proof.
elim : s => [| t s'] //.
move => IH Hu.
rewrite set_cons.
move/setP => H.
move : (eq_card H).
rewrite cards0 cardsU1 inE.
rewrite cons_uniq in Hu.
by move:Hu; move/andP => [-> _].
Qed.


Lemma InME : ex_inclusion_mx = seqmx_of_mx _ InM.
Proof.
apply/seqmxP; split=> [|j0|j0 k].
    by rewrite size_map /Vector.dim /= mul1n -(n_simplicesE tr_leT irr_leT n all_sorted_i).
  rewrite /Vector.dim /= mul1n -(n_simplicesE tr_leT irr_leT n all_sorted_i) => Hi.
  by rewrite /rowseqmx (nth_map [::]) // !size_map mul1n -(n_simplicesE tr_leT irr_leT n all_sorted_ip).
rewrite mxE; move: j0 k; rewrite /fun_of_seqmx /rowseqmx.
rewrite /Vector.dim /= !mul1n -(n_simplicesE tr_leT irr_leT n all_sorted_ip) -(n_simplicesE tr_leT irr_leT n all_sorted_i) => j0 k.
rewrite !(nth_map [::]) //.
case:ifP=> H.
  rewrite (eqP H); case : ifP => //.
  by move/eqP.
case : ifP => //.
move => H1.
have H2 : forall x, x \in [set t in nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0]
-> x \in [set t in nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) k] by move => x1 H3; rewrite -(eqP H1) H3.
move : H; move/eqP => H.
move : H1.
case H1 : (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) => [| t s].
  have -> : [set t in nil] = set0 by done.
  move : (uniq_ip (ltn_ord k)) => H2'.
  rewrite eq_sym; move/eqP => H3.
  move : (set0_setin H2' H3).
  rewrite -{2}H1 => H4.
  by move : H; rewrite -H4.
have Hsize :
  size (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0) =
  size (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) k)
  by rewrite (size_i (ltn_ord j0)) (size_ip (ltn_ord k)).
have Hs1 :
  sorted leT (nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0)
  by exact : (sorted_i (ltn_ord j0)).
have Hs2 :
  sorted leT
    (nth [::] (ex_n_simplices (nth [::] filtration_faces p) n) k)
  by exact : (sorted_ip (ltn_ord k)).
move : (sorted_exists Hsize Hs1 Hs2 H) => [x [Hx1 Hx2]].
move/eqP/setP => H1'.
have Hx1': x \in [set t in nth [::] (ex_n_simplices (nth [::] filtration_faces j) n) j0] by rewrite inE.
move : (H2 x Hx1').
rewrite inE => H3.
by move : Hx2; rewrite H3.
Qed.

End executable_persistent_inclusion_mx.

Section executable_persistent_betti.

Local Open Scope ring_scope.

Variable V : finType.
Variable f : seq (seq (seq V)).
Variable n j p: nat.


Definition ex_p_persistent_n_betti_K_j :=
 ex_PBetti_rank
    (size (ex_n_simplices (nth nil f j) n))
    (size (ex_n_1_simplices (nth nil f j) n))
    (size (ex_n_simplices (nth nil f p) n.+1))
    (size (ex_n_simplices (nth nil f p) n))
    (trseqmx (ex_incidence_mx (nth nil f j) n))
    (trseqmx (ex_incidence_mx (nth nil f p) n.+1))
    (ex_inclusion_mx f n j p).

End executable_persistent_betti.

End PersistenceFiltration.


(*
(* Some examples *)

Local Open Scope ring_scope.

Definition filtration_triangle : seq (seq (seq 'I_3)) :=
[::[::[::0]];
   [::[::0];[::1]];
   [::[::0];[::1];[::@Ordinal 3 2 (refl_equal true)]];
   [::[::0;1];[::@Ordinal 3 2 (refl_equal true)]];
   [::[::0;1];[::1;@Ordinal 3 2 (refl_equal true)]];
   [::[::0;1];[::1;@Ordinal 3 2 (refl_equal true)];[::0;@Ordinal 3 2 (refl_equal true)]];
   [::[::0;1;@Ordinal 3 2 (refl_equal true)]]].

Definition f := filtration_triangle.

Eval vm_compute in ex_p_persistent_k_betti_K_i f 1 5 6.
*)